from unittest import mock
from uuid import uuid4

from django.conf import settings
from django.test import TestCase
from jsm_user_services.services.user import current_jwt_token
from jsm_user_services.services.user import get_cpf_from_jwt
from jsm_user_services.services.user import get_jsm_token
from jsm_user_services.services.user import get_jsm_user_data_from_jwt
from jsm_user_services.services.user import get_ltm_token
from jsm_user_services.services.user import get_user_email_from_jwt
from jsm_user_services.services.user import get_user_ip
from jsm_user_services.services.user import is_retail_user
from jsm_user_services.services.user import jwt_has_required_roles
from jsm_user_services.support.local_threading_utils import add_to_local_threading
from jsm_user_services.support.local_threading_utils import remove_from_local_threading
from requests_mock import Mocker


class TestContracts(TestCase):
    def setUp(self):

        self.jwt = "eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.eyJ0b2tlbl90eXBlIjoiYWNjZXNzIiwiZXhwIjoxNTYxODAzNDAzLCJqdGkiOiI0MDJlYmJjNDdkNGE0NDUxYmRjZDIwMmQyYTEwOTdiMiIsInl1bnRpYW5kdSI6InhJRTVucGZ4d3l0dXZhQVlWRW5rdUxDclVhcWl0YW9xY09kMUZvZXJCT045TW5qdmNyQ3YzWXZnYXVibk9IMEVVR2swbkl6NmppTWNZdlphckJGZ2M4clVycVVjN3NKcGVfVWZieklPZjNldXlTbTZ3Wmc2SEpQU2d2dmN2akhNTnVTcVR6ZmduUzhUV3ZWNTYyQmI5UHFEUUZjODdUdS0zR0gzdEpJWTRZQ3NEYjRJRXFXLUJSVUJSYlpPRGZ6NnNLSkxGekZEaVRMa2djWGZ6MVJ6b1c5ZGtOYXVUWnZ5Wkk1dkhOeWJhTzRwLVlnWmJ0YkZXb1ktMWFCZ0x4MzFUeGdNUmxsYS1UZzBkU2dveGM1OTV2d3U5aDlnTUh3cGVVNU5DQ0gyS2g0NnZmNGExRlVzM3I5TjlWYUJzQlBFRkpwR2RhNkdzS0l2a1RKMHd5QTZLOUN3OUtJYS04Sy15bUFCUmdHWFVaSTN3LWMzdnB0Xzd2TkQzbFhEZC1CU0xVdWhubDd3VE92QnYxNGpjS1gweGR2MzJmRUdXam1WSnNIa3QxTDl2NG9PRmhmTWJJeEFoY2RLc1IyTUlDemVZZjh0ODJJZDhSMEZqNUJMV3BQVXM1VG5OQXV3UzJOcm1UQWtfT0psZUk1TGdBRlBjVEpXcWZDeDl5cDJab2FMT1FteGF6aEdCNHo5MnRBamFQYTVySEFaWTBibThBbVZvakxjVkIwMWVJeUVwN3dPY2VFRGI0UVRkdEtoREZFcGxOa1ZHakRQNTZLbDYzRThWTHpoemhXQllQRXdPd2RHS3FKT09Lek4wcDN1NWhtQWZJSTFHUHFHb0ZIaTNFbEpNRlB0TnhxN0R0LUo4VGpHSGo1NEc0QWRqUUVvNFpqdjBPNFBtS0R6SmRUeTNGM0JiQjZMU092czVfd0V1bkFlSkJqREFSOW1jQWdNcWxsbDVDRXJhNjc4cmQzMUE1Vm51MXl3WXFuLTlfUSIsImpzbV9pZGVudGl0eSI6ImV5SmhiR2NpT2lKSVV6STFOaUlzSW5SNWNDSTZJa3BYVkNKOS5leUpsYldGcGJDSTZJakV5TXpRMVFHeHZlV0ZzZEhsS1UwMHVZMjl0SWl3aWRXNXBjWFZsWDI1aGJXVWlPaUpoYzJRaUxDSjNaV0p6YVhSbElqb2lRMnhwWlc1MFFYQndiR2xqWVhScGIyNGlMQ0oxYVdRaU9pSTROVEEzTVdSa1pTMWlOVFk0TFRReFlXWXRZVFUzT1MwelpqUXlZekZrWmpBNU9UZ2lMQ0p6Wld4c1pYSnpJanBiSWpGbE0ySTRNekZtTFRrMVpEQXROREZoTnkxaU0yVXhMVEZtWlRBelptRmxNakEwTmlJc0lqWTNaV1k0TXpBeUxXSTFORFl0TkRRMVlTMDVZV0l5TFRobU4yWTRPR0kyTWpBM09DSmRMQ0p5YjJ4bElqb2lRM1Z6ZEc5dFpYSWlMQ0pqYm5CcWN5STZJalk0TGpVNU55NHlPREF2TURBd01TMHdPU0lzSW01aVppSTZNVFUyTVRjek1UUXdNaXdpWlhod0lqb3hOVFl5TXpNMk1qQXlMQ0pwWVhRaU9qRTFOakUzTXpFME1ESXNJbWx6Y3lJNkluUm1jRG92TDJsa1pXNTBhWFI1TG1wemJTOWhkWFJvYjNKcGRIa2lMQ0poZFdRaU9pSmlZekExTVdSbE15MWhaV0ZoTFRSaFl6a3RPR1ppWmkwd05XRTRObVV3T0RObVkyWWlmUS5sRHpDdWFQeTJobDhJUDlFb3p0QmpwR2NLVkh5RVZLdVgxWUtMeUgxd0pBIn0.CzhdhYN6porjALfSktQvE9gIfK34BHibFOLjZ-BUAzw"
        self.user_ip = "201.192.124.34"

        settings

    def test_current_jwt_token_should_return_none_when_local_threading_dont_have_key(self):
        remove_from_local_threading("authorization_token")

        self.assertIsNone(current_jwt_token())

    def test_current_jwt_token_should_return_none_when_local_threading_have_key(self):
        add_to_local_threading("authorization_token", self.jwt)

        self.assertIsNotNone(current_jwt_token())
        self.assertEqual(current_jwt_token(), self.jwt)

    def test_get_jsm_token_should_return_none_when_current_jwt_token_return_none(self):
        remove_from_local_threading("authorization_token")

        self.assertIsNone(get_jsm_token())

    def test_get_jsm_token_should_return_token_when_current_jwt_token_have_token(self):

        token_jsm = "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJlbWFpbCI6IjEyMzQ1QGxveWFsdHlKU00uY29tIiwidW5pcXVlX25hbWUiOiJhc2QiLCJ3ZWJzaXRlIjoiQ2xpZW50QXBwbGljYXRpb24iLCJ1aWQiOiI4NTA3MWRkZS1iNTY4LTQxYWYtYTU3OS0zZjQyYzFkZjA5OTgiLCJzZWxsZXJzIjpbIjFlM2I4MzFmLTk1ZDAtNDFhNy1iM2UxLTFmZTAzZmFlMjA0NiIsIjY3ZWY4MzAyLWI1NDYtNDQ1YS05YWIyLThmN2Y4OGI2MjA3OCJdLCJyb2xlIjoiQ3VzdG9tZXIiLCJjbnBqcyI6IjY4LjU5Ny4yODAvMDAwMS0wOSIsIm5iZiI6MTU2MTczMTQwMiwiZXhwIjoxNTYyMzM2MjAyLCJpYXQiOjE1NjE3MzE0MDIsImlzcyI6InRmcDovL2lkZW50aXR5LmpzbS9hdXRob3JpdHkiLCJhdWQiOiJiYzA1MWRlMy1hZWFhLTRhYzktOGZiZi0wNWE4NmUwODNmY2YifQ.lDzCuaPy2hl8IP9EoztBjpGcKVHyEVKuX1YKLyH1wJA"

        add_to_local_threading("authorization_token", self.jwt)

        self.assertIsNotNone(get_jsm_token())
        self.assertEqual(get_jsm_token(), token_jsm)

    def test_get_ltm_token_should_return_none_when_current_jwt_token_return_none(self):
        remove_from_local_threading("authorization_token")

        self.assertIsNone(get_ltm_token())

    def test_get_ltm_token_token_should_return_token_when_current_jwt_token_have_token(self):

        token_ltm = "xIE5npfxwytuvaAYVEnkuLCrUaqitaoqcOd1FoerBON9MnjvcrCv3YvgaubnOH0EUGk0nIz6jiMcYvZarBFgc8rUrqUc7sJpe_UfbzIOf3euySm6wZg6HJPSgvvcvjHMNuSqTzfgnS8TWvV562Bb9PqDQFc87Tu-3GH3tJIY4YCsDb4IEqW-BRUBRbZODfz6sKJLFzFDiTLkgcXfz1RzoW9dkNauTZvyZI5vHNybaO4p-YgZbtbFWoY-1aBgLx31TxgMRlla-Tg0dSgoxc595vwu9h9gMHwpeU5NCCH2Kh46vf4a1FUs3r9N9VaBsBPEFJpGda6GsKIvkTJ0wyA6K9Cw9KIa-8K-ymABRgGXUZI3w-c3vpt_7vND3lXDd-BSLUuhnl7wTOvBv14jcKX0xdv32fEGWjmVJsHkt1L9v4oOFhfMbIxAhcdKsR2MICzeYf8t82Id8R0Fj5BLWpPUs5TnNAuwS2NrmTAk_OJleI5LgAFPcTJWqfCx9yp2ZoaLOQmxazhGB4z92tAjaPa5rHAZY0bm8AmVojLcVB01eIyEp7wOceEDb4QTdtKhDFEplNkVGjDP56Kl63E8VLzhzhWBYPEwOwdGKqJOOKzN0p3u5hmAfII1GPqGoFHi3ElJMFPtNxq7Dt-J8TjGHj54G4AdjQEo4Zjv0O4PmKDzJdTy3F3BbB6LSOvs5_wEunAeJBjDAR9mcAgMqlll5CEra678rd31A5Vnu1ywYqn-9_Q"

        add_to_local_threading("authorization_token", self.jwt)

        self.assertIsNotNone(get_ltm_token())
        self.assertEqual(get_ltm_token(), token_ltm)

    def test_get_jsm_user_data_from_jwt_should_return_none_when_local_threading_dont_have_key(self):
        remove_from_local_threading("authorization_token")

        self.assertIsNone(get_jsm_user_data_from_jwt())

    def test_get_jsm_user_data_from_jwt_should_return_data_when_local_threading_have_key(self):

        data = {
            "email": "12345@loyaltyJSM.com",
            "unique_name": "asd",
            "website": "ClientApplication",
            "uid": "85071dde-b568-41af-a579-3f42c1df0998",
            "sellers": ["1e3b831f-95d0-41a7-b3e1-1fe03fae2046", "67ef8302-b546-445a-9ab2-8f7f88b62078"],
            "role": "Customer",
            "cnpjs": "68.597.280/0001-09",
            "nbf": 1561731402,
            "exp": 1562336202,
            "iat": 1561731402,
            "iss": "tfp://identity.jsm/authority",
            "aud": "bc051de3-aeaa-4ac9-8fbf-05a86e083fcf",
        }

        add_to_local_threading("authorization_token", self.jwt)

        self.assertIsNotNone(get_jsm_user_data_from_jwt())
        self.assertEqual(get_jsm_user_data_from_jwt(), data)

    def test_get_email_from_jwt_should_return_email_when_token_exists(self):
        add_to_local_threading("authorization_token", self.jwt)

        self.assertEqual(get_user_email_from_jwt(), "12345@loyaltyJSM.com")

    def test_get_cpf_from_jwt(self):
        add_to_local_threading("authorization_token", self.jwt)

        self.assertEqual(get_cpf_from_jwt(), "12345")

    @mock.patch("jsm_user_services.services.user.get_jsm_user_data_from_jwt")
    def test_jwt_has_required_roles_assert_all_true(self, mocked_get_jsm_user_data_from_jwt):
        mocked_get_jsm_user_data_from_jwt.return_value = {
            "iss": "Online JWT Builder",
            "iat": 1559177717,
            "exp": 1590713717,
            "aud": "www.example.com",
            "sub": "jrocket@example.com",
            "email": "1234567890@email.com",
            "roles": ["Dev", "Project Administrator"],
        }

        self.assertTrue(jwt_has_required_roles(["Dev", "Project Administrator"]))

    @mock.patch("jsm_user_services.services.user.get_jsm_user_data_from_jwt")
    def test_jwt_has_required_roles_assert_all_false(self, mocked_get_jsm_user_data_from_jwt):
        mocked_get_jsm_user_data_from_jwt.return_value = {
            "iss": "Online JWT Builder",
            "iat": 1559177717,
            "exp": 1590713717,
            "aud": "www.example.com",
            "sub": "jrocket@example.com",
            "email": "1234567890@email.com",
            "roles": ["Dev", "Project Administrator"],
        }

        self.assertFalse(jwt_has_required_roles(["Dev"]))

    @mock.patch("jsm_user_services.services.user.get_jsm_user_data_from_jwt")
    def test_jwt_has_required_roles_assert_any_true(self, mocked_get_jsm_user_data_from_jwt):
        mocked_get_jsm_user_data_from_jwt.return_value = {
            "iss": "Online JWT Builder",
            "iat": 1559177717,
            "exp": 1590713717,
            "aud": "www.example.com",
            "sub": "jrocket@example.com",
            "email": "1234567890@email.com",
            "roles": ["Dev", "Project Administrator"],
        }

        self.assertTrue(jwt_has_required_roles(["Dev"], assert_all=False))

    @mock.patch("jsm_user_services.services.user.get_jsm_user_data_from_jwt")
    def test_jwt_has_required_roles_assert_any_false(self, mocked_get_jsm_user_data_from_jwt):
        mocked_get_jsm_user_data_from_jwt.return_value = {
            "iss": "Online JWT Builder",
            "iat": 1559177717,
            "exp": 1590713717,
            "aud": "www.example.com",
            "sub": "jrocket@example.com",
            "email": "1234567890@email.com",
            "roles": ["Dev", "Project Administrator"],
        }

        self.assertFalse(jwt_has_required_roles(["X"], assert_all=False))

    def test_should_return_false_when_user_is_not_retail_user(self):
        user_id_ref = str(uuid4())

        with Mocker(real_http=True) as mocker:
            mocked_request = mocker.get(
                f"{settings.USER_API_HOST}/users/search/?user_id_ref={user_id_ref}&is_retail_user=True",
                status_code=200,
                json={"count": 0},
            )

            self.assertFalse(is_retail_user(user_id_ref))
            self.assertTrue(mocked_request.called_once)

    def test_should_return_true_when_user_is_retail_user(self):
        user_id_ref = str(uuid4())

        with Mocker(real_http=True) as mocker:
            mocked_request = mocker.get(
                f"{settings.USER_API_HOST}/users/search/?user_id_ref={user_id_ref}&is_retail_user=True",
                status_code=200,
                json={"count": 1},
            )

            self.assertTrue(is_retail_user(user_id_ref))
            self.assertTrue(mocked_request.called_once)

    def test_get_user_ip_should_return_none_when_local_threading_dont_have_key(self):
        remove_from_local_threading("user_ip")

        self.assertIsNone(get_user_ip())

    def test_get_user_ip_should_not_return_none_when_local_threading_have_key(self):
        add_to_local_threading("user_ip", self.user_ip)

        self.assertIsNotNone(get_user_ip())
        self.assertEqual(get_user_ip(), self.user_ip)



"""
HTTP utils module
"""
import functools
import logging
from contextlib import contextmanager
from typing import Any
from typing import Generator

from django.conf import settings
from requests import Response
from requests import Session
from requests.adapters import HTTPAdapter
from urllib3 import Retry

JSM_USER_SERVICE_HTTP_TIMEOUT = int(getattr(settings, "JSM_USER_SERVICE_HTTP_TIMEOUT", 30))


def _log_response(response: Response, *args: Any, **kwargs: Any) -> None:
    logging.debug("Request to %s returned status code %d", response.url, response.status_code)


def _check_for_errors(response: Response, *args: Any, **kwargs: Any) -> None:
    response.raise_for_status()


@contextmanager
def request(total: int = 3, backoff_factor: float = 0.1, **kwargs: Any) -> Generator[Session, None, None]:
    """
    Generate a requests session that allows retries and raises HTTP errors (status code >= 400).
    Uses the same arguments as the class Retry from urllib3
    """
    session = Session()
    session.hooks.update(response=[_log_response, _check_for_errors])
    max_retries = Retry(total=total, backoff_factor=backoff_factor, **kwargs)
    adapter = HTTPAdapter(max_retries=max_retries)
    session.mount("https://", adapter)
    session.mount("http://", adapter)

    session.request = functools.partial(session.request, timeout=JSM_USER_SERVICE_HTTP_TIMEOUT)
    try:
        yield session
    finally:
        session.close()


def get_response_body(response: Response) -> Any:
    """
    Deserialize response
    """
    try:
        response_body = response.json()
    except ValueError:
        response_body = response.text
    return response_body



import logging
import re
from typing import Optional

from django.http import HttpRequest
from jsm_user_services.support.local_threading_utils import add_to_local_threading
from jsm_user_services.support.local_threading_utils import remove_from_local_threading

try:
    from django.utils.deprecation import MiddlewareMixin
except ImportError:
    MiddlewareMixin = object

logger = logging.getLogger(__name__)


class JsmJwtService(MiddlewareMixin):
    def process_request(self, request):
        add_to_local_threading("authorization_token", self._get_jwt_token_from_request(request))
        add_to_local_threading("user_ip", self._get_user_ip_from_request(request))

    def process_response(self, request, response):
        remove_from_local_threading("authorization_token")
        remove_from_local_threading("user_ip")
        return response

    @staticmethod
    def _get_jwt_token_from_request(request: HttpRequest) -> Optional[str]:
        """
        Extracts JWT token from a Django request object.
        """
        authorization_token = request.META.get("HTTP_AUTHORIZATION", "")

        match = re.match("Bearer", authorization_token)

        if not match:
            return None

        auth_type_beginning = match.span()[1]
        jwt_token = authorization_token[auth_type_beginning:].strip()

        return jwt_token

    @staticmethod
    def _get_user_ip_from_request(request: HttpRequest) -> Optional[str]:
        """
        Retrieve the user ip that made this request from Django HttpRequest object
        """
        return request.META.get("X-Original-Forwarded-For", None)



from unittest import TestCase

from jsm_user_services.support.http_utils import request
from requests.exceptions import ConnectTimeout


class TestContracts(TestCase):
    def test_should_raise_timeout(self):

        with request() as r:
            # check https://stackoverflow.com/a/100859 for the reason of "http://www.google.com:81/"
            self.assertRaises(ConnectTimeout, r.get, "http://www.google.com:81/")



"""
Module with tests for the DRF tools.
"""
from unittest import mock

import pytest
from django.test import TestCase
from requests.exceptions import HTTPError

from jsm_user_services.drf_tools.permissions import RoleBasedPermission
from jsm_user_services.drf_tools.permissions import StatusBasedPermission


class TestDRFPermissionClasses(TestCase):
    def setUp(self):
        self.role_based_permission = RoleBasedPermission()
        self.status_based_permission = StatusBasedPermission()

    @mock.patch("jsm_user_services.drf_tools.permissions.get_user_data_from_server")
    def test_role_based_permission_should_validate_request_data_and_append_it_to_request(
        self, mocked_get_user_data_from_server
    ):
        request = mock.MagicMock()
        del request.user_data
        append_user_data_to_request = True
        remote_user_data = {"name": "Igor G Peternella", "roles": ["professional"]}

        # get_user_data_from_server answers with an http 200 status
        mocked_get_user_data_from_server.return_value = remote_user_data
        user_data = self.role_based_permission._validate_request_against_user_service(
            request, append_user_data_to_request
        )

        expected_user_data = remote_user_data

        self.assertDictEqual(user_data, expected_user_data)
        self.assertDictEqual(request.user_data, expected_user_data)

    @mock.patch("jsm_user_services.drf_tools.permissions.get_user_data_from_server")
    def test_role_based_permission_should_return_empty_dict_upon_401_from_user_service(
        self, mocked_get_user_data_from_server
    ):
        request = mock.MagicMock()
        del request.user_data
        append_user_data_to_request = True

        # get_user_data_from_server answers with an http 401 status
        error = HTTPError("Unauthorized error!")
        response = mock.MagicMock()
        response.status_code = 401
        setattr(error, "response", response)

        mocked_get_user_data_from_server.side_effect = error

        user_data = self.role_based_permission._validate_request_against_user_service(
            request, append_user_data_to_request
        )

        expected_user_data = {}

        self.assertDictEqual(user_data, expected_user_data)

    @mock.patch("jsm_user_services.drf_tools.permissions.get_user_data_from_server")
    def test_role_based_permission_should_reraise_exception_when_user_service_returns_500(
        self, mocked_get_user_data_from_server
    ):
        request = mock.MagicMock()
        del request.user_data
        append_user_data_to_request = True

        # get_user_data_from_server answers with an http 500 status
        error = HTTPError("A 500 was returned!")
        response = mock.MagicMock()
        response.status_code = 500
        setattr(error, "response", response)

        mocked_get_user_data_from_server.side_effect = error

        with pytest.raises(HTTPError) as e:
            self.role_based_permission._validate_request_against_user_service(request, append_user_data_to_request)
            assert str(e.value) == "A 500 was returned!"

    @mock.patch("jsm_user_services.drf_tools.permissions.get_user_data_from_server")
    def test_permission_shouldnt_call_get_user_data_from_server_when_user_data_is_already_appended_to_request(
        self, mocked_get_user_data_from_server
    ):
        request = mock.MagicMock()
        remote_user_data = {"name": "Remote User", "roles": ["professional"]}
        request.user_data = remote_user_data

        user_data = self.role_based_permission._validate_request_against_user_service(request)

        self.assertDictEqual(user_data, remote_user_data)
        mocked_get_user_data_from_server.assert_not_called()

    @mock.patch("jsm_user_services.drf_tools.permissions.RoleBasedPermission._validate_request_against_user_service")
    def test_validate_user_role_should_return_false_if_user_data_is_an_empty_dict(
        self, mocked___validate_request_against_user_service
    ):
        request = mock.MagicMock()
        mocked___validate_request_against_user_service.return_value = {}

        permission_rslt = self.role_based_permission._validate_user_role(request, allowed_roles=["professional"])
        self.assertFalse(permission_rslt)

    @mock.patch("jsm_user_services.drf_tools.permissions.RoleBasedPermission._validate_request_against_user_service")
    def test_validate_user_role_should_return_false_if_user_data_has_disallowed_role(
        self, mocked___validate_request_against_user_service
    ):
        request = mock.MagicMock()
        mocked___validate_request_against_user_service.return_value = {
            "roles": ["owner"]  # only 'professional' role is allowed
        }

        permission_rslt = self.role_based_permission._validate_user_role(request, allowed_roles=["professional"])
        self.assertFalse(permission_rslt)

    @mock.patch("jsm_user_services.drf_tools.permissions.RoleBasedPermission._validate_request_against_user_service")
    def test_validate_user_role_should_return_true_if_user_data_has_allowed_role(
        self, mocked___validate_request_against_user_service
    ):
        request = mock.MagicMock()
        mocked___validate_request_against_user_service.return_value = {
            "roles": ["professional"]  # only 'professional' role is allowed
        }

        permission_rslt = self.role_based_permission._validate_user_role(request, allowed_roles=["professional"])
        self.assertTrue(permission_rslt)

    @mock.patch("jsm_user_services.drf_tools.permissions.RoleBasedPermission._validate_request_against_user_service")
    def test_validate_user_role_should_return_true_if_user_data_has_allowed_role_when_more_than_one_role_is_allowed(
        self, mocked___validate_request_against_user_service
    ):
        request = mock.MagicMock()
        mocked___validate_request_against_user_service.return_value = {
            "roles": ["manager"]  # only 'manager' and 'employee' roles are allowed
        }

        permission_rslt = self.role_based_permission._validate_user_role(request, allowed_roles=["employee", "manager"])
        self.assertTrue(permission_rslt)

    @mock.patch("jsm_user_services.drf_tools.permissions.StatusBasedPermission._validate_request_against_user_service")
    def test_validate_user_status_should_return_true_if_user_data_has_allowed_status(
        self, mocked___validate_request_against_user_service
    ):
        request = mock.MagicMock()
        mocked___validate_request_against_user_service.return_value = {
            "status": "active"
        }  # only 'active' status is allowed

        permission_rslt = self.status_based_permission._validate_user_status(request, allowed_status=["active"])
        self.assertTrue(permission_rslt)

    @mock.patch("jsm_user_services.drf_tools.permissions.StatusBasedPermission._validate_request_against_user_service")
    def test_validate_user_status_should_return_true_if_user_data_has_allowed_status_when_more_than_one_status_is_allowed(
        self, mocked___validate_request_against_user_service
    ):
        request = mock.MagicMock()
        mocked___validate_request_against_user_service.return_value = {
            "status": "pending-validation"
        }  # only 'active' and 'pending-validation' status are allowed

        permission_rslt = self.status_based_permission._validate_user_status(
            request, allowed_status=["active", "pending-validation"]
        )
        self.assertTrue(permission_rslt)

    @mock.patch("jsm_user_services.drf_tools.permissions.StatusBasedPermission._validate_request_against_user_service")
    def test_validate_user_status_should_return_false_if_user_data_has_disallowed_status(
        self, mocked___validate_request_against_user_service
    ):
        request = mock.MagicMock()
        mocked___validate_request_against_user_service.return_value = {
            "status": "blocked"
        }  # only pending-validation' status is allowed

        permission_rslt = self.status_based_permission._validate_user_status(
            request, allowed_status=["pending-validation"]
        )
        self.assertFalse(permission_rslt)

    @mock.patch("jsm_user_services.drf_tools.permissions.StatusBasedPermission._validate_request_against_user_service")
    def test_validate_user_status_should_return_false_if_user_data_is_an_empty_dict(
        self, mocked___validate_request_against_user_service
    ):
        request = mock.MagicMock()
        mocked___validate_request_against_user_service.return_value = {}

        permission_rslt = self.status_based_permission._validate_user_status(request, allowed_status=["active"])
        self.assertFalse(permission_rslt)



"""
Module with useful user permissions classes meant to be used with Django Rest Framework.
"""
from typing import Any
from typing import Dict
from typing import List

import requests
from django.conf import settings
from rest_framework import permissions
from rest_framework.request import Request

from jsm_user_services.services.user import get_user_data_from_server


class JSMUserBasePermission(permissions.BasePermission):
    """
    Base class for JSM user permissions. Implements methods for validating requests against
    the user micro service so that any type of permission can be carried out (role, status, etc).
    """

    APPEND_USER_DATA = getattr(settings, "JSM_USER_SERVICES_DRF_APPEND_USER_DATA", True)
    USER_DATA_ATTR_NAME = getattr(settings, "JSM_USER_SERVICES_DRF_REQUEST_USER_DATA_ATTR_NAME", "user_data")
    USER_SERVICE_NOT_AUTHORIZED_CODES = (401, 403, 404)

    @classmethod
    def _retrieve_user_data(cls, request: Request) -> Dict:
        try:
            user_data = getattr(request, cls.USER_DATA_ATTR_NAME)
        except AttributeError:
            user_data = get_user_data_from_server()

        return user_data

    @classmethod
    def _validate_request_against_user_service(cls, request: Request, append_user_data_to_request: bool = True) -> Dict:
        """
        Gets valid user_data from the User micro service.
        """
        try:
            user_data: Dict = cls._retrieve_user_data(request)

            if append_user_data_to_request:
                setattr(request, cls.USER_DATA_ATTR_NAME, user_data)

        except requests.exceptions.HTTPError as e:
            if e.response.status_code in cls.USER_SERVICE_NOT_AUTHORIZED_CODES:
                return {}

            raise

        return user_data

    def has_permission(self, request: Request, view: Any) -> bool:
        """
        Abstract method that must be implemented by children classes.
        """
        raise NotImplementedError("This method must be overridden!")


class StatusBasedPermission(JSMUserBasePermission):
    """
    Base class for the status-based permissions. Implements methods for validating requests against
    the user micro service and an utility method that asserts that the user has the appropriate status.
    """

    @classmethod
    def _validate_user_status(cls, request: Request, allowed_status: List[str]):
        """
        Validates if the user has the appropriate status against the allowed status.
        """
        append_user_data_to_request = cls.APPEND_USER_DATA
        user_data = cls._validate_request_against_user_service(
            request, append_user_data_to_request=append_user_data_to_request
        )

        if "status" not in user_data:
            return False

        return user_data["status"] in allowed_status


class RoleBasedPermission(JSMUserBasePermission):
    """
    Base class for the role-based permissions. Implements methods for validating requests against
    the user micro service and an utility method that asserts that the user has the appropriate role.
    """

    @classmethod
    def _validate_user_role(cls, request: Request, allowed_roles: List[str]):
        """
        Validates if the user has the appropriate role against the allowed roles.
        """
        append_user_data_to_request = cls.APPEND_USER_DATA
        user_data = cls._validate_request_against_user_service(
            request, append_user_data_to_request=append_user_data_to_request
        )
        user_roles: List[str] = user_data.get("roles", [])

        if not user_roles:
            return False

        return any([user_role in allowed_roles for user_role in user_roles])


class ProfessionalUserPermission(RoleBasedPermission):
    """
    Permission that allows only users with the role 'professional'.
    """

    def has_permission(self, request: Request, view: Any) -> bool:
        """
        Asserts that the user is a professional.
        """
        return self._validate_user_role(request, allowed_roles=["professional"])


class EmployeeOrManagerUserPermission(RoleBasedPermission):
    """
    Permission that allows only users with the role 'employee' or 'manager'.
    """

    def has_permission(self, request: Request, view: Any) -> bool:
        """
        Asserts that the user is either an employee or manager.
        """
        return self._validate_user_role(request, allowed_roles=["employee", "manager"])


class EmployeeOnlyUserPermission(RoleBasedPermission):
    """
    Permission that allows only users with the role 'employee'.
    """

    def has_permission(self, request: Request, view: Any) -> bool:
        """
        Asserts that the user is an employee.
        """
        return self._validate_user_role(request, allowed_roles=["employee"])


class OwnerUserPermission(RoleBasedPermission):
    """
    Permission that allows only users with the role 'owner'.
    """

    def has_permission(self, request: Request, view: Any) -> bool:
        """
        Asserts that the user is an owner.
        """
        return self._validate_user_role(request, allowed_roles=["owner"])


class AnyLoggedUserPermission(RoleBasedPermission):
    """
    Permission that allows users to access the requested resource regardless
    of their roles as long as they have a VALID, NON EXPIRED JSM JWT.
    """

    def has_permission(self, request: Request, view: Any) -> bool:
        """
        Asserts that the user is just logged in with a valid jwt (non expired).
        """
        return self._validate_user_role(request, allowed_roles=["professional", "employee", "manager", "owner"])


class ActiveUserPermission(StatusBasedPermission):
    """
    Permission that allows only users which the status is 'active'.
    """

    def has_permission(self, request: Request, view: Any) -> bool:
        """
        Asserts that the user status is 'active'.
        """
        return self._validate_user_status(request, allowed_status=["active"])


class PendingValidationUserPermission(StatusBasedPermission):
    """
    Permission that allows only users which the status is 'pending-validation'.
    """

    def has_permission(self, request: Request, view: Any) -> bool:
        """
        Asserts that the user status is 'pending-validation'.
        """
        return self._validate_user_status(request, allowed_status=["pending-validation"])
